/*
 * Copyright (C) 2004-2010 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file ogg_afh.c Audio format handler for ogg/vorbis files. */

#include <vorbis/codec.h>
#include <regex.h>

#include "para.h"
#include "afh.h"
#include "error.h"
#include "string.h"
#include "ogg_afh_common.h"

struct private_vorbis_data {
	vorbis_info vi;
	vorbis_comment vc;
};

static int vorbis_packet_callback(ogg_packet *packet, int packet_num,
		struct afh_info *afhi, void *private_data)
{
	struct private_vorbis_data *pvd = private_data;

	if (vorbis_synthesis_headerin(&pvd->vi, &pvd->vc, packet) < 0)
		return -E_VORBIS;
	if (packet_num == 0) {
		if (pvd->vi.rate == 0)
			return -E_VORBIS;
		afhi->channels = pvd->vi.channels;
		afhi->frequency = pvd->vi.rate;
		afhi->bitrate = pvd->vi.bitrate_nominal / 1000;
		PARA_DEBUG_LOG("channels: %i, sampling rate: %i, bitrate: %i\n",
			afhi->channels, afhi->frequency, afhi->bitrate);
		return 1;
	}
	if (packet_num == 1)
		return 1; /* we also want to have packet #2 */
	afhi->tags.artist = para_strdup(vorbis_comment_query(&pvd->vc, "artist", 0));
	afhi->tags.title = para_strdup(vorbis_comment_query(&pvd->vc, "title", 0));
	afhi->tags.album = para_strdup(vorbis_comment_query(&pvd->vc, "album", 0));
	afhi->tags.year = para_strdup(vorbis_comment_query(&pvd->vc, "year", 0));
	afhi->tags.comment = para_strdup(vorbis_comment_query(&pvd->vc, "comment", 0));
	return 0;
}

static int ogg_vorbis_get_file_info(char *map, size_t numbytes, __a_unused int fd,
		struct afh_info *afhi)
{
	int ret;
	struct private_vorbis_data pvd;
	struct ogg_afh_callback_info vorbis_callback_info = {
		.packet_callback = vorbis_packet_callback,
		.private_data = &pvd,
	};

	vorbis_info_init(&pvd.vi);
	vorbis_comment_init(&pvd.vc);
	ret = ogg_get_file_info(map, numbytes, afhi, &vorbis_callback_info);
	vorbis_info_clear(&pvd.vi);
	vorbis_comment_clear(&pvd.vc);
	return ret;
}

static const char* ogg_suffixes[] = {"ogg", NULL};

/**
 * The init function of the ogg vorbis audio format handler.
 *
 * \param afh Pointer to the struct to initialize.
 */
void ogg_init(struct audio_format_handler *afh)
{
	afh->get_file_info = ogg_vorbis_get_file_info,
	afh->suffixes = ogg_suffixes;
}
