/*
 * Copyright (C) 2006-2010 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file sched.c Paraslash's scheduling functions. */

#include <regex.h>
#include <dirent.h> /* readdir() */
#include <assert.h>
#include <sys/time.h>

#include "para.h"
#include "ipc.h"
#include "fd.h"
#include "list.h"
#include "sched.h"
#include "string.h"
#include "time.h"
#include "error.h"

static struct list_head pre_select_list, post_select_list;
static int initialized;

static struct timeval now_struct;
struct timeval *now = &now_struct;

/*
 * Remove a task from the scheduler.
 *
 * \param t The task to remove.
 *
 * If the pre_select pointer of \a t is not \p NULL, it is removed from
 * the pre_select list of the scheduler. Same goes for \a post_select.
 */
static void unregister_task(struct task *t)
{
	if (!initialized)
		return;
	PARA_INFO_LOG("unregistering %s (%s)\n", t->status,
		t->error <0? para_strerror(-t->error) : "shutdown");
	if (t->pre_select)
		list_del(&t->pre_select_node);
	if (t->post_select)
		list_del(&t->post_select_node);
	t->error = -E_TASK_UNREGISTERED;
}


static void sched_preselect(struct sched *s)
{
	struct task *t, *tmp;
	list_for_each_entry_safe(t, tmp, &pre_select_list, pre_select_node) {
		if (t->error >= 0 && t->pre_select)
			t->pre_select(s, t);
//		PARA_INFO_LOG("%s \n", t->status);
		if (t->error >= 0)
			continue;
		/*
		 * We have to check whether the list is empty because the call
		 * to ->pre_select() might have called sched_shutdown(). In
		 * this case t has been unregistered already, so we must not
		 * unregister it again.
		 */
		if (list_empty(&pre_select_list))
			return;
		unregister_task(t);
	}
}

//#define SCHED_DEBUG 1
static inline void call_post_select(struct sched *s, struct task *t)
{
#ifndef SCHED_DEBUG
	t->post_select(s, t);
#else
	struct timeval t1, t2, diff;
	unsigned long pst;

	gettimeofday(&t1, NULL);
	t->post_select(s, t);
	gettimeofday(&t2, NULL);
	tv_diff(&t1, &t2, &diff);
	pst = tv2ms(&diff);
	if (pst > 50)
		PARA_WARNING_LOG("%s: post_select time: %lums\n",
			t->status, pst);
#endif
}

static void sched_post_select(struct sched *s)
{
	struct task *t, *tmp;

	list_for_each_entry_safe(t, tmp, &post_select_list, post_select_node) {
		if (t->error >= 0)
			call_post_select(s, t);
//		PARA_INFO_LOG("%s: %d\n", t->status, t->ret);
		if (t->error >= 0)
			continue;
		/* nec., see sched_preselect() */
		if (list_empty(&post_select_list))
			return;
		unregister_task(t);
	}
}

/**
 * The core function for all paraslash programs.
 *
 * \param s Pointer to the scheduler struct.
 *
 * This function updates the global \a now pointer, calls all registered
 * pre_select hooks which may set the timeout and add any file descriptors to
 * the fd sets of \a s.  Next, it calls para_select() and makes the result available
 * to the registered tasks by calling their post_select hook.
 *
 * \return Zero if no more tasks are left in either of the two lists, negative
 * if para_select returned an error.
 *
 * \sa task, now.
 */
int schedule(struct sched *s)
{
	int ret;

	if (!initialized)
		return -E_NOT_INITIALIZED;
	if (!s->select_function)
		s->select_function = para_select;
again:
	FD_ZERO(&s->rfds);
	FD_ZERO(&s->wfds);
	s->select_timeout = s->default_timeout;
	s->max_fileno = -1;
	gettimeofday(now, NULL);
	sched_preselect(s);
	if (list_empty(&pre_select_list) && list_empty(&post_select_list))
		return 0;
	ret = s->select_function(s->max_fileno + 1, &s->rfds, &s->wfds,
		&s->select_timeout);
	if (ret < 0)
		return ret;
	if (ret == 0) {
		/*
		 * APUE: Be careful not to check the descriptor sets on return
		 * unless the return value is greater than zero. The return
		 * state of the descriptor sets is implementation dependent if
		 * either a signal is caught or the timer expires.
		 */
		FD_ZERO(&s->rfds);
		FD_ZERO(&s->wfds);
	}
	gettimeofday(now, NULL);
	sched_post_select(s);
	if (list_empty(&pre_select_list) && list_empty(&post_select_list))
		return 0;
	goto again;
}

/*
 * Initialize the paraslash scheduler.
 */
static void init_sched(void)
{
	PARA_INFO_LOG("initializing scheduler\n");
	INIT_LIST_HEAD(&pre_select_list);
	INIT_LIST_HEAD(&post_select_list);
	initialized = 1;
}

/**
 * Add a task to the scheduler.
 *
 * \param t the task to add
 *
 * If the pre_select pointer of \a t is not \p NULL, it is added to
 * the pre_select list of the scheduler. Same goes for post_select.
 *
 * \sa task::pre_select, task::post_select
 */
void register_task(struct task *t)
{
	if (!initialized)
		init_sched();
	PARA_INFO_LOG("registering %s (%p)\n", t->status, t);
	if (t->pre_select) {
		PARA_DEBUG_LOG("pre_select: %p\n", &t->pre_select);
		list_add_tail(&t->pre_select_node, &pre_select_list);
	}
	if (t->post_select) {
		PARA_DEBUG_LOG("post_select: %p\n", &t->post_select);
		list_add_tail(&t->post_select_node, &post_select_list);
	}
}

/**
 * Unregister all tasks.
 *
 * This will cause \a schedule() to return immediately because both the
 * \a pre_select_list and the \a post_select_list are empty.
 */
void sched_shutdown(void)
{
	struct task *t, *tmp;

	if (!initialized)
		return;
	list_for_each_entry_safe(t, tmp, &pre_select_list, pre_select_node)
		unregister_task(t);
	list_for_each_entry_safe(t, tmp, &post_select_list, post_select_node)
		unregister_task(t);
	initialized = 0;
}

/**
 * Get the list of all registered tasks.
 *
 * \return The task list.
 *
 * Each entry of the list contains an identifier which is simply a hex number
 * that may be used in \a kill_task() to terminate the task.
 * The result is dynamically allocated and must be freed by the caller.
 */
char *get_task_list(void)
{
	struct task *t, *tmp;
	char *msg = NULL;

	if (!initialized)
		return NULL;
	list_for_each_entry_safe(t, tmp, &pre_select_list, pre_select_node) {
		char *tmp_msg;
		tmp_msg = make_message("%s%p\tpre\t%s\n", msg? msg : "", t, t->status);
		free(msg);
		msg = tmp_msg;
	}
	list_for_each_entry_safe(t, tmp, &post_select_list, post_select_node) {
		char *tmp_msg;
//		if (t->pre_select)
//			continue;
		tmp_msg = make_message("%s%p\tpost\t%s\n", msg? msg : "", t, t->status);
		free(msg);
		msg = tmp_msg;
	}
	//PARA_DEBUG_LOG("task list:\n%s", msg);
	return msg;
}

/**
 * Simulate an error for the given task.
 *
 * \param id The task identifier.
 *
 * Find the task identified by \a id, set the tasks' error value to
 * \p -E_TASK_KILLED and unregister the task.
 *
 * \return Positive on success, negative on errors (e.g. if \a id does not
 * correspond to a registered task).
 */
int kill_task(char *id)
{
	struct task *t, *tmp;
	char buf[20];

	if (!initialized)
		return -E_NOT_INITIALIZED;
	list_for_each_entry_safe(t, tmp, &pre_select_list, pre_select_node) {
		sprintf(buf, "%p", t);
		if (strcmp(id, buf))
			continue;
		t->error = -E_TASK_KILLED;
		return 1;
	}
	list_for_each_entry_safe(t, tmp, &post_select_list, post_select_node) {
		sprintf(buf, "%p", t);
		if (strcmp(id, buf))
			continue;
		t->error = -E_TASK_KILLED;
		return 1;
	}
	return -E_NO_SUCH_TASK;
}

/**
 * Set the select timeout to the minimal possible value.
 *
 * \param s Pointer to the scheduler struct.
 *
 * This causes the next select() call to return immediately.
 */
void sched_min_delay(struct sched *s)
{
	s->select_timeout.tv_sec = 0;
	s->select_timeout.tv_usec = 1;
}

/**
 * Impose an upper bound for the timeout of the next select() call.
 *
 * \param to Maximal allowed timeout.
 * \param s Pointer to the scheduler struct.
 *
 * If the current scheduler timeout is already smaller than \a to, this
 * function does nothing. Otherwise the timeout for the next select() call is
 * set to the given value.
 *
 * \sa sched_request_timeout_ms().
 */
void sched_request_timeout(struct timeval *to, struct sched *s)
{
	if (tv_diff(&s->select_timeout, to, NULL) > 0)
		s->select_timeout = *to;
}

/**
 * Force the next select() call to return before the given amount of milliseconds.
 *
 * \param ms The maximal allowed timeout in milliseconds.
 * \param s Pointer to the scheduler struct.
 *
 * Like sched_request_timeout() this imposes an upper bound on the timeout
 * value for the next select() call.
 */
void sched_request_timeout_ms(long unsigned ms, struct sched *s)
{
	struct timeval tv;
	ms2tv(ms, &tv);
	sched_request_timeout(&tv, s);
}

/**
 * Force the next select() call to return before the given future time.
 *
 * \param barrier Absolute time before select() should return.
 * \param s Pointer to the scheduler struct.
 *
 * If \a barrier is in the past, this function does nothing.
 *
 * \sa sched_request_barrier_or_min_delay().
 */
void sched_request_barrier(struct timeval *barrier, struct sched *s)
{
	struct timeval diff;

	if (tv_diff(now, barrier, &diff) > 0)
		return;
	sched_request_timeout(&diff, s);
}

/**
 * Force the next select() call to return before the given time.
 *
 * \param barrier Absolute time before select() should return.
 * \param s Pointer to the scheduler struct.
 *
 * If \a barrier is in the past, this function requests a minimal timeout.
 *
 * \sa sched_min_delay(), sched_request_barrier().
 */
void sched_request_barrier_or_min_delay(struct timeval *barrier, struct sched *s)
{
	struct timeval diff;

	if (tv_diff(now, barrier, &diff) > 0)
		return sched_min_delay(s);
	sched_request_timeout(&diff, s);
}
